const std = @import("std");
const assert = std.debug.assert;
const parseInt = std.fmt.parseInt;

pub const Cpio = struct {
    const Self = @This();
    const end_marker = "TRAILER!!!";

    remaining_files: []const u8,

    pub fn init(cpio: []const u8) Cpio {
        return .{
            .remaining_files = cpio,
        };
    }

    pub fn findFile(self: *const Self, name: []const u8) ?(CpioError!CpioEntry_) {
        var copy = self.*;

        while (copy.nextFile()) |err_entry| {
            const entry = err_entry catch break;

            if (std.mem.eql(u8, entry.name() catch continue, name))
                return entry;
        }

        return null;
    }

    pub fn nextFile(self: *Self) ?(CpioError!CpioEntry_) { // *const CpioEntry) {
        if (self.remaining_files.len == 0)
            return null;

        if (self.remaining_files.len < 110) // @sizeOf(CpioEntry))
            return CpioError.EndOfData;

        // const entry = @ptrCast(*const CpioEntry, &self.remaining_files[0]);
        const entry = CpioEntry_{
            .header = self.remaining_files[0..110],
        };

        const size = try entry.total_size();

        if (size > self.remaining_files.len)
            return CpioError.EndOfData;

        self.remaining_files = self.remaining_files[size..];

        try entry.valid();

        if (std.mem.eql(u8, try entry.name(), end_marker))
            return null;

        return entry;
    }
};

pub const CpioError = error{
    EndOfData,
    Invalid,
    ChecksumMismatch,
};

const CpioEntry_ = struct {
    comptime {
        assert(@sizeOf(CpioEntry_) != 110); // Switch to CpioEntry
    }

    const Self = @This();
    const magic_value = [6]u8{ '0', '7', '0', '7', '0', '1' };

    // Fields are zero filled octals and a null byte
    header: []const u8,

    fn f_magic(self: *const Self) []const u8 {
        return self.header[0..6];
    }

    fn f_filesize(self: *const Self) []const u8 {
        return self.header[54..62];
    }

    fn f_namesize(self: *const Self) []const u8 {
        return self.header[94..102];
    }

    pub fn data(self: *const Self) CpioError![]const u8 {
        const filesize = try read_hex(self.f_filesize());
        return @intToPtr([*]const u8, try self.file_start())[0..filesize];
    }

    pub fn name(self: *const Self) CpioError![:0]const u8 {
        const size = try read_hex(self.f_namesize());

        return @intToPtr([*]const u8, @ptrToInt(&self.header[0]) + 110)[0 .. size - 1 :0];
    }

    pub fn valid(self: *const Self) CpioError!void {
        if (!std.mem.eql(u8, self.f_magic(), &magic_value))
            return CpioError.Invalid;
    }

    fn file_end(self: *const Self) CpioError!usize {
        const end_of_file = (try self.file_start()) + (try read_hex(self.f_filesize()));
        return end_of_file + 3 & ~@as(usize, 0b11); // Align to four bytes
    }

    fn file_start(self: *const Self) CpioError!usize {
        const end_of_name = @ptrToInt(&self.header[0]) + 110 + try read_hex(self.f_namesize());
        return end_of_name + 3 & ~@as(usize, 0b11); // Align to four bytes
    }

    fn read_hex(digits: []const u8) CpioError!u32 {
        return parseInt(u32, digits, 16) catch return CpioError.Invalid;
    }

    fn total_size(self: *const Self) CpioError!usize {
        return (try self.file_end()) - @ptrToInt(&self.header[0]);
    }
};

// New ASCII format
const CpioEntry = packed struct {
    comptime {
        assert(@sizeOf(@This()) == 110);
    }

    const Self = @This();
    const magic_value = [6]u8{ '0', '7', '0', '7', '0', '1' };

    // Fields are zero filled octals and a null byte
    magic: [6]u8,
    ino: [8]u8,
    mode: [8]u8,
    uid: [8]u8,
    gid: [8]u8,
    nlink: [8]u8,
    mtime: [8]u8,
    filesize: [8]u8,
    devmajor: [8]u8,
    devminor: [8]u8,
    rdevmajor: [8]u8,
    rdevminor: [8]u8,
    namesize: [8]u8,
    check: [8]u8,

    pub fn data(self: *const Self) CpioError![]const u8 {
        const filesize = try read_hex(self.filesize);
        return @intToPtr([*]const u8, try self.file_start())[0..filesize];
    }

    pub fn name(self: *const Self) CpioError![:0]const u8 {
        const size = try read_hex(self.namesize);

        return @intToPtr([*]const u8, @ptrToInt(self) + @sizeOf(Self))[0 .. size - 1 :0];
    }

    pub fn valid(self: *const Self) CpioError!void {
        // This check causes a compiler segfault
        // if (!std.mem.eql(u8, &self.magic, &magic_value))
        // return CpioError.Invalid;
    }

    fn file_start(self: *const Self) CpioError!usize {
        const end_of_name = @ptrToInt(self) + @sizeOf(Self) + try read_hex(self.namesize);
        return end_of_name + 3 & ~@as(usize, 0b11); // Align to four bytes
    }

    fn file_end(self: *const Self) CpioError!usize {
        const end_of_file = (try self.file_start()) + (try read_hex(self.filesize));
        return end_of_file + 3 & ~@as(usize, 0b11); // Align to four bytes
    }

    fn read_hex(digits: []u8) CpioError!u32 {
        return parseInt(u32, digits, 16) catch return CpioError.Invalid;
    }

    fn total_size(self: *const Self) CpioError!usize {
        return (try self.file_end()) - @ptrToInt(self);
    }
};

test "Size" {
    const x: CpioEntry_ = undefined;
}

test "Empty stream" {
    const stream = [_]u8{
        0x30, 0x37, 0x30, 0x37, 0x30, 0x31, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x31, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x30, 0x30, 0x30, 0x42, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,
        0x30, 0x30, 0x54, 0x52, 0x41, 0x49, 0x4c, 0x45, 0x52, 0x21, 0x21, 0x21,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    };

    var cpio = Cpio.init(&stream);

    const end = cpio.nextFile();
    std.debug.warn("\n{}\n", .{end});
    assert(end == null);
}
