const std = @import("std");

pub const Ppm = struct {
    const Self = @This();

    slice: []const u8,
    width: usize,
    height: usize,
    max_val: usize,

    pub fn init(slice: []const u8) !Self {
        var self = Self{
            .slice = slice,
            .width = 0,
            .height = 0,
            .max_val = 0,
        };

        const magic = self.consumeText();
        if (!std.mem.eql(u8, magic, "P6"))
            return Error.UnknownFormat;

        self.width = try self.consumeNumber();
        self.height = try self.consumeNumber();
        self.max_val = try self.consumeNumber();

        if (self.max_val != 255)
            return Error.UnsupportedFormat;

        if (self.slice.len == 0)
            return Error.InvalidFile;

        // Consume a single whitespace
        self.slice = self.slice[1..];

        if (self.slice.len != 3 * self.width * self.height)
            return Error.InvalidFile;

        return self;
    }

    pub fn nextPixel(self: *Self) ?RGBPixel {
        if (self.slice.len == 0)
            return null;

        const ret = RGBPixel{
            .r = self.slice[0],
            .g = self.slice[1],
            .b = self.slice[2],
        };

        self.slice = self.slice[3..];

        return ret;
    }

    fn consumeNumber(self: *Self) !usize {
        const num = self.consumeText();

        return std.fmt.parseInt(usize, num, 10);
    }

    fn consumeText(self: *Self) []const u8 {
        var start: usize = 0;
        while (start < self.slice.len and
            std.ascii.isSpace(self.slice[start]))
            start += 1;

        var end = start;
        while (end < self.slice.len and
            !std.ascii.isSpace(self.slice[end]))
            end += 1;

        const ret = self.slice[start..end];
        self.slice = self.slice[end..];
        return ret;
    }

    pub const RGBPixel = packed struct {
        r: u8,
        g: u8,
        b: u8,
    };

    pub const Error = error{
        InvalidFile,
        UnknownFormat,
        UnsupportedFormat,
    };
};

test "Empty file" {
    const assert = std.debug.assert;

    var ppm = try Ppm.init("P6 0 0 255 ");
    assert(ppm.width == 0);
    assert(ppm.height == 0);
    assert(ppm.max_val == 255);
    assert(ppm.nextPixel() == null);
}
