const std = @import("std");
const assert = std.debug.assert;

const cpu = @import("../cpu.zig");
const log = @import("../logger.zig").log;
const MicrocodeError = @import("../microcode.zig").MicrocodeError;
const msr = @import("../msr.zig");

// TODO Test
pub fn apply(microcode: []align(16) const u8) MicrocodeError!bool {
    var mc = try Microcode.init(microcode);
    const sign = cpu.CpuId.signature();
    const plat_id = cpu.rdmsr(msr.ia32_platform_id);

    if (mc.compatible(sign, plat_id)) {
        if (mc.header.loader_revision != 0x00000001)
            return MicrocodeError.Incompatible;

        log(.Debug, "[microcode/intel] Compatible microcode update found: {x}", .{mc.header.update_revision});

        cpu.wrmsr(msr.ia32_bios_updt_trig, @ptrToInt(&mc.microcode[0]));
    }

    return false;
}

pub fn parse(microcode: []const u8) MicrocodeError!Microcode {
    return try Microcode.init(microcode);
}

const Microcode = struct {
    const Self = @This();

    header: *align(16) const MicrocodeUpdateHeader,
    microcode: []align(16) const u8,
    processor_signatures: ?[]const ProcessorSignature,

    pub fn init(microcode: []align(16) const u8) MicrocodeError!Self {
        if (microcode.len < @sizeOf(MicrocodeUpdateHeader))
            return MicrocodeError.EndOfData;

        const header = @ptrCast(*align(16) const MicrocodeUpdateHeader, &microcode[0]);

        if (header.header_version != 0x00000001)
            return MicrocodeError.Incompatible;

        if (header.total_size > microcode.len)
            return MicrocodeError.EndOfData;

        {
            var checksum: u32 = 0;
            for (@bytesToSlice(u32, microcode)) |v|
                checksum +%= v;
            if (checksum != 0)
                return MicrocodeError.ChecksumMismatch;
        }

        const data_size = if (header.data_size == 0) 2000 else header.data_size;
        const microcode_end = @sizeOf(MicrocodeUpdateHeader) + data_size;

        var processor_signatures: ?[]const ProcessorSignature = null;
        if (header.total_size > microcode_end) {
            // A extended processor signature table is present
            const signature_table_header = @ptrCast(*align(1) const ExtendedProcessorSignatureTableHeader, &microcode[microcode_end]);

            var checksum: u32 = 0;
            for (@bytesToSlice(u32, microcode[microcode_end..])) |v|
                checksum +%= v;
            if (checksum != 0)
                return MicrocodeError.ChecksumMismatch;

            const table_start = microcode_end + @sizeOf(ExtendedProcessorSignatureTableHeader);
            const table_end = table_start + @sizeOf(ProcessorSignature) * signature_table_header.count;

            processor_signatures = @bytesToSlice(ProcessorSignature, microcode[table_start..table_end]);
        }

        return Self{
            .header = header,
            .microcode = microcode[@sizeOf(MicrocodeUpdateHeader)..microcode_end],
            .processor_signatures = processor_signatures,
        };
    }

    pub fn compatible(self: *const Self, signature: cpu.CpuId.Signature, platform_id: u64) bool {
        const flag = @intCast(u3, platform_id >> 50 & 0b111);

        if (signature.value == self.header.processor_signature) {
            return self.header.processor_flags & @as(u32, 1) << flag != 0;
        }

        if (self.processor_signatures) |signatures| {
            for (signatures) |extra_sig| {
                if (signature.value == extra_sig.processor_signature)
                    return extra_sig.processor_flags & @as(u32, 1) << flag != 0;
            }
        }

        return false;
    }
};

const MicrocodeUpdateHeader = packed struct {
    header_version: u32,
    update_revision: u32,
    date: u32,
    processor_signature: u32,
    checksum: u32,
    loader_revision: u32,
    processor_flags: u8,
    _reserved1: u8, // Not a u24 or [3]u8 due to compiler bug
    _reserved2: u8,
    _reserved3: u8,
    data_size: u32,
    total_size: u32,
    _reserved4: u32, // Not a [3]u32 due to compiler bug
    _reserved5: u32,
    _reserved6: u32,

    comptime {
        assert(@sizeOf(@This()) == 48);
    }
};

const ExtendedProcessorSignatureTableHeader = packed struct {
    count: u32,
    checksum: u32,
    _reserved1: u32, // Not a [3]u32 due to compiler bug
    _reserved2: u32,
    _reserved3: u32,

    comptime {
        assert(@sizeOf(@This()) == 20);
    }
};

const ProcessorSignature = packed struct {
    processor_signature: u32,
    processor_flags: u32,
    checksum: u32,

    comptime {
        assert(@sizeOf(@This()) == 12);
    }
};
