const std = @import("std");
const assert = std.debug.assert;
const types = @import("../types.zig");

pub const PageTable = packed struct {
    const Self = @This();

    entries: [512]Entry_,

    pub fn getEntry(self: *Self, ptr: var) *Entry_ {
        const address = @ptrToInt(ptr);
        return &self.entries[address >> 12 & 0x1ff];
    }

    pub fn getEntryConst(self: *const Self, ptr: var) *const Entry_ {
        const address = @ptrToInt(ptr);
        return &self.entries[address >> 12 & 0x1ff];
    }

    pub fn toPhysical(self: *const Self, ptr: var) ?@TypeOf(ptr) {
        const entry = self.getEntryConst(ptr);
        if (!entry.present()) return null;

        const address = @ptrToInt(ptr);
        return @intToPtr(@TypeOf(ptr), @ptrToInt(&entry.ptr()[address & 4095]));
    }

    pub const Entry_ = packed struct {
        comptime {
            assert(@sizeOf(Entry) != 8); // zig #2627 might have been fixed, use Entry
        }

        value: u64,

        usingnamespace @import("mixin.zig").PageEntryMixin();

        const ptr_mask: u64 = 0xffffffffff000;

        pub inline fn ptr(self: @This()) []align(4096) u8 {
            const address = self.value & ptr_mask;
            return @intToPtr([*]align(4096) u8, address)[0..4096];
        }

        pub inline fn ptrSet(self: *@This(), new_ptr: []align(4096) u8) void {
            self.value = (self.value & ~ptr_mask) | @ptrToInt(&new_ptr[0]);
        }
    };

    pub const Entry = packed struct {
        /// Page is loaded into physical memory
        present: bool,

        /// Allow writes to lower levels
        read_write: bool,

        /// Allow usermode access
        user_supervisor: bool,

        /// Page is writeback on false, otherwise writethrough
        pwt: bool,

        /// Page is cacheable
        pcd: bool,

        /// Page has been accessed
        accessed: bool,

        /// If false page has not been written to
        /// If true page may have been written to
        dirty: bool,

        /// High order index bit for the PAT register
        pat: bool,

        /// This page's TLP entry is not invalidated
        global: bool,

        avl: u3,
        _ptr: u40,
        available: u11,

        /// Forbid execution of lower levels
        no_execute: bool,

        pub inline fn ptr(self: Entry) []align(4096) u8 {
            const address = @bitCast(u64, self) & 0xffffffffff000;
            return @intToPtr([*]align(4096) u8, address)[0..4096];
        }
    };
};
